#!/usr/bin/env python

import string

# ---------------------------------------------------------------------------- #
#                                 APPLICATION                                  #
# ---------------------------------------------------------------------------- #
class Application:
    def __init__(self, show=False, neighbour=False, trace=False):
        self.name = None
        self.collect_trace = trace
        self.neighbour_discover = neighbour
        self.show_cam_image = show

    def str_info(self):
        if not self.name:
            raise BaseException("str_info() method on base-class Application")
        ret = []
        ret.append('ApplicationName = "%s"' % self.name)
        ret.append('Application.collectTraceInfo = %s' % 
                   ('true' if self.collect_trace else 'false'))
        ret.append('Application.neighbourDiscover = %s' % 
                   ('true' if self.neighbour_discover else 'false'))
        ret.append('Application.showCamImage = %s' % 
                   ('true' if self.show_cam_image else 'false'))
        return ret

class Tracker(Application):
    def __init__(self, ts):
        Application.__init__(self)
        self.sampling_time = ts
    def str_info(self):
        ret = Application.str_info(self)
        ret.append('Application.sampling_time = %.2f' % self.sampling_time)
        return ret

class TrackerDPF(Tracker):
    def __init__(self, ts, particles, gmm, trace_particles=False):
        Tracker.__init__(self, ts)
        self.name = 'WiseCameraDPF'
        self.p = particles
        self.gmm = gmm
        self.trace_p = trace_particles

    def str_info(self):
        ret = Tracker.str_info(self)
        ret.append('Application.n_particles = %d' % self.p)
        ret.append('Application.n_gmm_components = %d' % self.gmm)
        ret.append('Application.trace_particles = %s' % 
                   ('true' if self.trace_p else 'false'))
        ret.append('Application.particle_spreading_factor = 10') #TODO: change?
        return ret

# ---------------------------------------------------------------------------- #
#                               COMMUNICATION                                  #
# ---------------------------------------------------------------------------- #
class Radio:
    def __init__(self):
        self.radio_file = '../BasicRadio.txt'
        self.mode = 'IDEAL'
        self.collision = 2
        self.max_frame_size = 0
        self.frame_overhead = 0
        self.subdir = ''

    def set_subdir(self, subdir):
        self.subdir = subdir + '/' 

    def str_info(self):
        ret = []
        ret.append('Radio.RadioParametersFile = "%s"' \
                   % (self.subdir + self.radio_file))
        ret.append('Radio.mode = "%s"' % self.mode)
        ret.append('Radio.collisionModel = %d' % self.collision)
        ret.append('Radio.maxPhyFrameSize = %d' % self.max_frame_size)
        ret.append('Radio.phyFrameOverhead = %d' % self.frame_overhead)
        return ret

class Mac:
    def __init__(self, buffer_size=None):
        self.name = None
        self.packet_overhead = 0
        self.buffer_size = buffer_size
    
    def str_info(self):
        if not self.name:
            raise BaseException("str_info() method called on base-class Mac")
        ret = []
        ret.append('MACProtocolName = "%s"' % self.name)
        ret.append('MAC.macPacketOverhead = %d' % self.packet_overhead)
        if self.buffer_size:
            ret.append('MAC.macBufferSize = %d' % self.buffer_size)
        return ret

class MacTMAC(Mac):
    def __init__(self, buffer_size, max_tx_retries, listen_timeout):
        Mac.__init__(self, buffer_size)
        self.name = 'TMAC'
        self.max_tx_retries = max_tx_retries
        self.listen_timeout = listen_timeout

    def str_info(self):
        ret = Mac.str_info(self)
        ret.append('MAC.maxTxRetries = %d' % self.max_tx_retries)
        ret.append('MAC.listenTimeout = %d' % self.listen_timeout)
        ret.append('MAC.allowSinkSync = false') #TODO: change?
        return ret

class Routing:
    def __init__(self):
        self.packet_overhead = 0

    def str_info(self):
        return ['Routing.netDataFrameOverhead = %d' % self.packet_overhead]


class Communication:
    def __init__(self, radio, mac, routing):
        self.radio = radio
        self.mac = mac
        self.routing = routing

    def str_info(self):
        ret = []
        ret.append('### Communication: Radio (Physical Layer)   ###')
        ret += map(lambda i: 'Communication.' + i, self.radio.str_info())
        ret.append('### Communication: MAC (Datalink Layer)     ###')
        ret += map(lambda i: 'Communication.' + i, self.mac.str_info())
        ret.append('### Communication: Routing (Network Layer)  ###')
        ret += map(lambda i: 'Communication.' + i, self.routing.str_info())
        return ret

# ---------------------------------------------------------------------------- #
#                                   SENSOR                                     #
# ---------------------------------------------------------------------------- #
class Sensor:
    def __init__(self, trace=False):
        self.name = None
        self.collect_trace = trace

    def str_info(self):
        if not self.name:
            raise BaseException("str_info() method called on base-class Sensor")
        ret = []
        ret.append('SensorManagerName = "%s"' % self.name)
        ret.append('SensorManager.collectTraceInfo = %s' % 
                   ('true' if self.collect_trace else 'false'))
        return ret

class SensorCamera(Sensor):
    def __init__(self, fov_type='bounding_box', fov_bb=(10, 6)):
        Sensor.__init__(self)
        self.name = 'WiseCameraManager'
        self.fov_type = fov_type
        self.fov_bb_w = fov_bb[0]
        self.fov_bb_h = fov_bb[1]

    def get_fov_bb(self):
        return (self.fov_bb_w, self.fov_bb_h)

    def str_info(self):
        ret = Sensor.str_info(self)
        ret.append('SensorManager.fov_type = "%s"' % self.fov_type)
        ret.append('SensorManager.fov_bb_width = %f' % self.fov_bb_w)
        ret.append('SensorManager.fov_bb_height = %f' % self.fov_bb_h)
        return ret

# ---------------------------------------------------------------------------- #
#                                    NODE                                      #
# ---------------------------------------------------------------------------- #
class Resource:
    clock_drift = 0

    def str_info(self):
        return ['ResourceManager.sigmaCPUClockDrift = %f' % self.clock_drift]


class Node:
    def __init__(self, n_id, x, y, z=0, phi=0, theta=0, 
                 app=None, com=None, sensor=None, res=None):
        self.n_id = n_id
        self.x = x
        self.y = y
        self.z = z
        self.phi = phi
        self.theta = theta
        self.app = app
        self.com = com
        self.sensor = sensor
        self.res = res

    def str_position(self):
        s = ''
        s += 'SN.node[%u].xCoor = %.2f\n' % (self.n_id, self.x)
        s += 'SN.node[%u].yCoor = %.2f\n' % (self.n_id, self.y)
        s += 'SN.node[%u].zCoor = %.2f\n' % (self.n_id, self.z)
        s += 'SN.node[%u].phi = %.2f\n' % (self.n_id, self.phi)
        s += 'SN.node[%u].theta = %.2f\n' % (self.n_id, self.theta)
        return s

    def str_application(self, star=False):
        out = ''
        info = self.app.str_info()
        out += '# ==============================================\n'
        out += '# Application Layer \n' 
        out += '# ==============================================\n'
        out += self.__create_string(info, star)
        return out

    def str_communication(self, star=False):
        out = ''
        info = self.com.str_info()
        out += '# ==============================================\n'
        out += '# Communication Module \n' 
        out += '# ==============================================\n'
        out += self.__create_string(info, star)
        return out

    def str_sensor(self, star=False):
        out = ''
        info = self.sensor.str_info()
        out += '# ==============================================\n'
        out += '# Sensor Manager  \n' 
        out += '# ==============================================\n'
        out += self.__create_string(info, star)
        return out

    def str_resource(self, star=False):
        out = ''
        info = self.res.str_info()
        out += '# ==============================================\n'
        out += '# Resource Manager \n' 
        out += '# ==============================================\n'
        out += self.__create_string(info, star)
        return out

    def __str__(self):
        out = ''
        out += self.str_position()
        out += self.str_application()
        out += self.str_communication()
        out += self.str_sensor()
        out += self.str_resource()
        return out
        
    def __create_string(self, info, star):
        prefix = 'SN.node[**].' if star else ('SN.node[%u].' % self.n_id)
        info = map(lambda i: (prefix if i[0]!='#' else '') + i + '\n', info)
        return string.join(info, '')
        
# ---------------------------------------------------------------------------- #
#                               GLOBAL MODULES                                 #
# ---------------------------------------------------------------------------- #
class WirelessChannel:
    dummy = False
    dummy_trace = False
    sigma = 0
    bidirectional_sigma = 0
    pld0 = -172 # should be disk radius = 1000m
    trace = False

    def __str__(self):
        s = ''
        s += '# ==============================================\n'
        s += '# Wireless Communication Channel\n' 
        s += '# ==============================================\n'
        s += 'SN.dummy_communication = ' + ('true' if self.dummy else 'false') 
        s += '\n'
        s += 'SN.wirelessChannel.sigma = ' + str(self.sigma) + '\n'
        s += 'SN.wirelessChannel.bidirectionalSigma = ' 
        s +=  str(self.bidirectional_sigma) + '\n'
        s += 'SN.wirelessChannel.PLd0 = ' + str(self.pld0) + '\n'
        s += 'SN.wirelessChannel.collectTraceInfo = ' 
        s += ('true' if self.trace else 'false') + '\n'
        s += 'SN.dummyChannel.collectTraceInfo = '
        s += ('true' if self.dummy_trace else 'false')  + '\n'
        return s

class PhysicalProcess:
    def __init__(self, p_id, trace=False):
        self.p_id = p_id
        self.collect_trace = trace

    def str_info(self):
        ret = []
        ret.append('physicalProcess.collectTraceInfo = %s' % 
                   ('true' if self.collect_trace else 'false'))
        return ret

class MovingTarget(PhysicalProcess):
    def __init__(self, p_id, motion=None, p_init=(0, 0), bb=(1.0, 1.0), 
                 update_time=20.00, min_move=None, max_move=None, log=False, 
                 move_after=0.0, noise=None):
        PhysicalProcess.__init__(self, p_id)
        self.name = 'WiseMovingTarget'
        self.motion = motion
        self.x_init = p_init[0]
        self.y_init = p_init[1]
        self.bb_width = bb[0]
        self.bb_height = bb[1]
        self.update_time = update_time
        self.min_x = min_move[0] if min_move else None
        self.min_y = min_move[1] if min_move else None
        self.max_x = max_move[0] if max_move else None
        self.max_y = max_move[1] if max_move else None
        self.log = log
        self.move_after = move_after
        self.motion_noise = noise
        self.rng_mapping = None

    def str_info(self):
        ret = PhysicalProcess.str_info(self)
        if not self.motion:
            self.motion = 'lin_noise'
        ret.append('move_type = "%s"' % self.motion)
        ret.append('x_init = %f' % self.x_init)
        ret.append('y_init = %f' % self.y_init)
        ret.append('bb_width = %f' % self.bb_width)
        ret.append('bb_height = %f' % self.bb_height)
        ret.append('update_time = %f' % self.update_time)
        ret.append('move_later = %f' % self.move_after)
        if self.motion_noise:
            ret.append('noise = %f' % self.motion_noise)
        if self.min_x:
            ret.append('move_min_x = %f' % self.min_x)
        if self.min_y:
            ret.append('move_min_y = %f' % self.min_y)
        if self.max_x:
            ret.append('move_max_x = %f' % self.max_x)
        if self.max_y:
            ret.append('move_max_y = %f' % self.max_y)
        if self.rng_mapping:
            for n, m in self.rng_mapping.items():
                # NOTE: 11 generators (0-10) are used in Castalia.ini
                ret.append('rng-%d = %d' % (n, m + 11))
        ret.append('log_enabled = %s' % ('true' if self.log else 'false'))
        return ret

class PhysicalProcessFactory:
    def __init__(self, proc_name, n_procs):
        self.name = proc_name
        self.n_procs = n_procs
        self.procs = []
        self.terrain_gui = ''
        self.terrain_gui_wait_start = False
        self.registry = {'WiseMovingTarget':MovingTarget}
        self.star = True
        if proc_name not in self.registry:
            raise BaseException("Invalid process name: valid values are ",
                                self.registry.keys())
        for p_id in range(n_procs):
            self.procs.append(self.registry[proc_name](p_id))

    def get_process(self, p_id):
        if p_id > self.n_procs or p_id < 0:
            raise BaseException("Invalid process index ", p_id)
        self.star = False
        return self.procs[p_id]
        
    def str_info(self):
        ret = []
        ret.append('SN.wiseTerrain.gui = "%s"' % self.terrain_gui)
        ret.append('SN.wiseTerrain.gui_wait_start = %s' % 
                   ('true' if self.terrain_gui_wait_start else 'false'))
        ret.append('SN.physicalProcessName = "%s"' % self.name)
        ret.append('###  ---  ###')
        for pos, p in enumerate(self.procs):
            info = p.str_info()
            if self.star and pos == 0:
                pre = 'SN.physicalProcess[**].'
            elif self.star and pos > 0:
                break;
            else:
                pre = 'SN.physicalProcess[%u].' % p.p_id
            info = map(lambda i: (pre if i[0]!='#' else '') + i, info)
            ret += info
        return ret

    def str_processes(self):
        s = ''
        s += '# ==============================================\n'
        s += '# Physical Process\n' 
        s += '# ==============================================\n'
        info = self.str_info()
        info = map(lambda i: i + '\n', info)
        s += string.join(info, '')
        return s

    def __str__(self):
        return self.str_processes()


# ---------------------------------------------------------------------------- #
#                               SENSOR NETWORK                                 #
# ---------------------------------------------------------------------------- #
class SensorNetwork:
    def __init__(self, w_channel, proc_factory, nodes, sizes=(100, 60, None), 
                 time_limit=None, rng_config=None):
        self.w_channel = w_channel
        self.field_x = sizes[0]
        self.field_y = sizes[1]
        self.field_z = sizes[2]
        self.time_limit = time_limit
        self.same_config = {}
        self.nodes = nodes
        self.proc_factory = proc_factory
        self.wise_base_path = '../../'
        self.subdir = ''
        self.rng = rng_config

    def set_subdir(self, subdir):
        self.subdir = subdir + '/'
        for n in self.nodes:
            if n.com:
               n.com.radio.set_subdir(subdir)

    def set_same_flag(self, tag, flag):
        valid_tags = ['app', 'com', 'sens', 'res']
        if tag not in valid_tags:
            msg = 'Tag "' + tag + '" is not valid for set_same_flag() method. '
            msg += 'Valid tags are: ' + str(valid_tags)
            raise BaseException(msg)
        if type(flag) != bool:
            raise BaseException('Invalid flag type, should be bool')
        self.same_config[tag] = flag

    def __str__(self):
        s = self.__header()
        s += self.__network_info()
        s += str(self.w_channel)
        s += '\n'
        #s += self.proc_factory.str_processes()
        s += str(self.proc_factory)
        s += '\n'
        pos = ''
        app  = ''
        sens = '' 
        res = ''
        com = ''
        for (i, n) in enumerate(self.nodes):
            pos += n.str_position()
            star = self.__check_same('app')
            if not star or (star and i == 0):
                app += n.str_application(star)
            star = self.__check_same('sens')
            if not star or (star and i == 0):
                sens += n.str_sensor(star)
            star = self.__check_same('res')
            if not star or (star and i == 0):
                res += n.str_resource(star)
            star = self.__check_same('com')
            if not star or (star and i == 0):
                com += n.str_communication(star)
        pos = '# ==============================================\n' + \
              "# Node Positions \n" + \
              '# ==============================================\n' +\
              pos
        s += pos + app + sens + res + com
        return s

    def __check_same(self, key):
        return (key in self.same_config and self.same_config[key])

    def __header(self):
        s = ''
        s += '# ####################################################### #\n'
        s += '# Description: this was generated                         #\n'
        s += '# ####################################################### #\n'
        s += '\n'
        s += '[General]\n'
        path = self.subdir + self.wise_base_path
        s += 'include ' + path + '../Simulations/Parameters/Castalia.ini\n'
        s += 'ned-path = ' + path + '../src/;' + path + 'src/\n'
        s += '\n'
        if self.time_limit:
            s += 'sim-time-limit = %fs\n' % self.time_limit
        if self.rng:
            # NOTE: 11 generators (0-10) are used in Castalia.ini
            s += 'num-rngs = %d\n' % (len(self.rng) + 11) 
            for idx, seed in self.rng.items():
                if seed != None:
                    s += 'seed-%d-mt = %d\n' % (idx + 11, seed)
        return s

    def __network_info(self):
        s = ''
        s += '# ==============================================\n'
        s += '# Sensor Network setup\n'
        s += '# ==============================================\n'
        s += 'SN.field_x = %f       # meters\n' % self.field_x
        s += 'SN.field_y = %f       # meters\n' % self.field_y
        if self.field_z:
            s += 'SN.field_z = %f       # meters\n' % self.field_z
        s += 'SN.numNodes = %d \n' % len(self.nodes)
        s += 'SN.numPhysicalProcesses = %d \n' % self.proc_factory.n_procs
        s += '\n'
        return s
