#!/usr/bin/env python

import os
from scipy import stats as scipy_stats

class Value():
    pass

################################################################################
#                                Plot Class                                    #
################################################################################
class Plot():
    def __init__(self, eps, ftag, xlabel, ylabel, plot_err, plot_maxmin):
        self.out_eps = eps
        self.file_tag = ftag
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.plot_err = plot_err
        self.plot_maxmin = plot_maxmin

    def do_plot(self, data):
        self.__gnuplot(data)
    
    def __gnuplot(self, data):
        script = 'plotter_' + self.file_tag + '.gpi'
        d_info, err, mm = self.__gnuplot_datafiles(data)
        err = err and self.plot_err
        mm = mm and self.plot_maxmin
        self.__gnuplot_script(d_info, script, err, mm)
        os.system('gnuplot ' + script)
        
    def __gnuplot_datafiles(self, info, prefix='plot_'):
        new_info = []
        has_err = False
        has_min = False
        has_max = False
        for (name, data) in info:
            data_fname = prefix + self.file_tag + '_' + name + '.dat'
            with open(data_fname, 'w') as f:
                for v in data:
                    s = str(v.n_nodes) + ' '
                    s += str(v.val) + ' '
                    if v.sem != None:
                        has_err = True
                        s += str(v.sem / 2) + ' '
                    if v.min != None:
                        has_min = True
                        s += str(v.min) + ' '
                    if v.max != None:
                        has_max = True
                        s += str(v.max) + ' '
                    print >> f, s
            new_info.append((name, data_fname))
        return new_info, has_err, (has_min and has_max)

    def __gnuplot_script(self, info, script, err, minmax):
        with open(script, 'w') as f:
            print >> f, 'set term postscript eps dashed color '\
                        'blacktext "Helvetica" 12'
            print >> f, 'set grid'
            if self.ylabel:
                print >> f, 'set ylabel "%s"' % self.ylabel
            if self.xlabel:
                print >> f, 'set xlabel "%s"' % self.xlabel
            print >> f, 'set output "%s"' % self.out_eps
            s = 'plot '
            for i, (label, fname) in enumerate(info):
                if err:
                    s += '"' + fname + '" '
                    s += 'using 1:2:3 with yerrorbars title "" '
                    s += 'pt 0 lt 1 lw 1 lc %d, ' % (i + 1)
                if minmax:
                    pass
                    s += '"' + fname + '" '
                    s += 'using 1:4 with points title "" '
                    s += 'pt 9 lc %d, ' % (i + 1)
                    s += '"' + fname + '" '
                    s += 'using 1:5 with points title "" '
                    s += 'pt 11 lc %d, ' % (i + 1)
                s += '"' + fname + '" '
                s += 'using 1:2 with linespoints '
                s += 'pt 2 lt 1 lw 2 lc %d ' % (i + 1)
                s += 'title "' + label + '"'
                s += ',' if i < len(info) - 1 else ';'
            print >> f, s

################################################################################
#                            Statistics Base Class                             #
################################################################################
class stat():
    def __init__(self, data):
        self.data = data
        self.statistics = {}
        self.plot = None

    def calculate(self, records):
        raise BaseException('calculate() method called on base class')

    def calculate_tid(self, tid):
        self.statistics[tid] = []
        for runner, results in self.data:
            info = []
            for n_nodes, res in results:
                if tid not in res.stats:
                    continue
                v = self.calculate(res.stats[tid])
                v.n_nodes = n_nodes
                info.append(v)
            self.statistics[tid].append((runner.name, info))
        return self.statistics[tid] 

    def plot_tid(self, tid, out_eps=None, plot_maxmin=False):
        if not self.plot:
            raise BaseException('Plot object not ready. Have you created stat?')
        if tid not in self.statistics:
            self.calculate_tid(tid)
        if out_eps:
            self.plot.out_eps = out_eps
        if plot_maxmin:
            self.plot.plot_maxmin = plot_maxmin
        self.plot.do_plot(self.statistics[tid])

################################################################################
#                             Statistics Classes                               #
################################################################################
class mse(stat):
    def __init__(self, data):
        stat.__init__(self, data)
        self.plot = Plot('plot_mse.eps', 'mse', '#Nodes', 'MSE', True, False)

    def calculate(self, records):
        valids = filter(lambda r: not r.t_lost, records)
        valids = map(lambda r: r.error, valids)
        v = Value()
        v.val = scipy_stats.tmean(valids)
        v.min = scipy_stats.tmin(valids)
        v.max = scipy_stats.tmax(valids, None)
        v.sem = scipy_stats.tsem(valids)
        return v

class mse2(stat):
    def __init__(self, data):
        stat.__init__(self, data)
        self.plot = Plot('plot_mse2.eps', 'mse2', '#Nodes', 'MSE2', True, False)

    def calculate(self, records):
        #valids = filter(lambda r: not r.t_lost, records)
        valids = map(lambda r: r.error, records)
        v = Value()
        v.val = scipy_stats.tmean(valids)
        v.min = scipy_stats.tmin(valids)
        v.max = scipy_stats.tmax(valids, None)
        v.sem = scipy_stats.tsem(valids)
        return v

class mse3(stat):
    def __init__(self, data):
        stat.__init__(self, data)
        self.plot = Plot('plot_mse3.eps', 'mse3', '#Nodes', 'MSE3', True, False)

    def calculate(self, records):
        valids = filter(lambda r: not r.t_lost, records)
        valids = map(lambda r: r.error, valids)
        k = 1.0 / float(len(valids))
        v = Value()
        v.val = k * scipy_stats.tmean(valids)
        v.min = k * scipy_stats.tmin(valids)
        v.max = k * scipy_stats.tmax(valids, None)
        v.sem = k * scipy_stats.tsem(valids)
        return v

class delay(stat):
    def __init__(self, data):
        stat.__init__(self, data)
        self.out_eps = 'plot_delay.eps'
        self.file_tag = 'delay'
        self.plot = Plot('plot_delay.eps', 'delay', '#Nodes', 'Avg Delay', 
                         True, False)
        
    def calculate(self, records):
        valids = filter(lambda r: not r.t_lost, records)
        valids = map(lambda r: r.delay_a, valids)
        v = Value()
        v.val = scipy_stats.tmean(valids)
        v.min = scipy_stats.tmin(valids)
        v.max = scipy_stats.tmax(valids, None)
        v.sem = scipy_stats.tsem(valids)
        return v
        
class samples(stat):
    def __init__(self, data):
        stat.__init__(self, data)
        self.plot = Plot('plot_procSamples.eps', 'samples', '#Nodes', 
                         '#Samples', True, False)
        
    def calculate(self, records):
        valids = filter(lambda r: not r.t_lost, records)
        v = Value()
        v.val = len(valids)
        v.min = None
        v.max = None
        v.sem = None
        return v

class other(stat):
    def __init__(self, data):
        stat.__init__(self, data)
        self.plot = Plot('plot_procSamples.eps', 'samples', '#Nodes', 
                         '#Samples', True, False)
        
    def calculate(self, records):
        samples = map(lambda r: r.t_lost, records)
        lost = samples[0]
        print 'test ', samples
        print 'whose length was ', len(samples)
        print 'init state is ', lost
        counted = []
        cnt = 0
        for s in samples:
            if lost == True and s != lost:
                counted.append(cnt)
                cnt = 0
            elif (lost == True and s == lost) or (lost == False and s != lost):
                cnt += 1
            lost = s
        tot = sum(counted)
        counted = map(lambda c: (float(c) / len(samples)) * 100, counted)
        print 'listone = ', counted
        print 'totale = ', tot
        v = Value()
        v.val = scipy_stats.tmean(counted)
        v.min = scipy_stats.tmin(counted)
        v.max = scipy_stats.tmax(counted, None)
        v.sem = scipy_stats.tsem(counted)
        print 'media = ', v.val 
        print 'min = ', v.min
        print 'max = ', v.max
        print 'sem = ', v.sem
        return v
