// *****************************************************************************
//  Copyright (C): Christian Nastasi, 2011                                    
//  Author(s): Christian Nastasi                                              
//  Developed at the:                                                         
//  - Retis LAB, TeCIP Institute, Scuola Superiore Sant'Anna (Pisa)           
//  - School of Elec Eng and Computer Science, Queen Mary University (London) 
//  This file is distributed under the terms in the attached LICENSE_2 file.  
//  If you do not find this file, copies can be found by writing to:          
//  - c.nastasi@sssup.it                                                      
//  - nastasichr@gmail.com                                                    
//  - andrea.cavallaro@eecs.qmul.ac.uk                                        
// *****************************************************************************

#include "WiseBaseApplication.h"
#include "WiseDebug.h"

#define perr_app(m) opp_error("\n[Application]:\n "m)
#define perr(m) opp_error(m)

void WiseBaseApplication::initialize()
{
	cMessage *m;
	readParameter();

	/* Send the STARTUP message to: 
	 * 1) Sensor_Manager;
	 * 2) Commmunication module;
	 * 3) APP (self message) so that the node starts to operate.
	 * We are using a random delay IF we must not have node synchronized
	 */
	double startup_delay = simTime().dbl();
	if (!synchronizedStartup)
		startup_delay += genk_dblrand(0) * 0.05 * cpuClockDrift;
	m = new cMessage("Communication [STARTUP]", NODE_STARTUP);
	sendDelayed(m, startup_delay, "toCommunicationModule");
	m = new cMessage("Sensor Dev Mgr [STARTUP]", NODE_STARTUP);
	sendDelayed(m, startup_delay, "toSensorDeviceManager");
	m = new cMessage("App [STARTUP]", NODE_STARTUP);
	scheduleAt(startup_delay, m);

	neighbourDiscover(startup_delay);

	/* Latency measurement is optional. An application can not define the 
	 * following two parameters. If they are not defined then the
	 * declareHistogram and collectHistogram statement are not called.
	 */
	latencyMax = hasPar("latencyHistogramMax") ? 
		     par("latencyHistogramMax") : 0;
	latencyBuckets = hasPar("latencyHistogramBuckets") ? 
			 par("latencyHistogramBuckets") : 0;
	if (latencyMax > 0 && latencyBuckets > 0)
		declareHistogram("Application level latency, in ms", 0, 
				 latencyMax, latencyBuckets);
	last_packet_latency = -1;
}

void WiseBaseApplication::handleMessage(cMessage * msg)
{
	int msgKind = msg->getKind();
	WISE_DEBUG_3("WiseBaseApplication::handleMessage()");

	if (disabled && msgKind != NODE_STARTUP) {
		delete msg;
		msg = NULL;	// safeguard
		return;
	}

	switch (msgKind) {
	case NODE_STARTUP:
		disabled = 0;
		startup();
		break;
	case APPLICATION_PACKET:
		WISE_DEBUG_3("\tApplication Packet");
		WiseApplicationPacket *rcvPacket;
		rcvPacket = check_and_cast <WiseApplicationPacket*>(msg);
		processAppPacket(rcvPacket);
		break;
	case TIMER_SERVICE:
		handleTimerMessage(msg);
		break;
	case SENSOR_READING_MESSAGE:
		WISE_DEBUG_3("\tSENSOR_READING_MESSAGE");
		WiseSensorManagerMessage *sensMsg;
		sensMsg = check_and_cast <WiseSensorManagerMessage*>(msg);
		handleSensorReading(sensMsg);
		break;
	case OUT_OF_ENERGY:
		disabled = 1;
		break;
	case DESTROY_NODE:
		disabled = 1;
		break;
	case NETWORK_CONTROL_MESSAGE:
		WISE_DEBUG_3("\tNetwork Control");
		handleNetworkControlMessage(msg);
		break;
	case MAC_CONTROL_MESSAGE:
		WISE_DEBUG_3("\tMac Control" << endl);
		handleMacControlMessage(msg);
		break;
	case RADIO_CONTROL_MESSAGE:
		WISE_DEBUG_3("\tRadio Control" << endl);
		RadioControlMessage *radioMsg;
		radioMsg = check_and_cast <RadioControlMessage*>(msg);
		handleRadioControlMessage(radioMsg);
		break;
	default:
		perr("Application module received unexpected message");
	}

	delete msg;
	msg = NULL;		// safeguard
}

void WiseBaseApplication::finish()
{
	CastaliaModule::finish();
	DebugInfoWriter::closeStream();
}

//ApplicationGenericDataPacket *WiseBaseApplication::createGenericDataPacket(double data, int seqNum, int size)
//{
//	ApplicationGenericDataPacket *newPacket =
//	    new ApplicationGenericDataPacket("Application generic packet", APPLICATION_PACKET);
//	newPacket->setData(data);
//	newPacket->setSequenceNumber(seqNum);
//	if (size < 0)
//		size = constantDataPayload;
//	if (size < 0)
//		perr("cannot create generic data packet of negative size");
//	size += packetHeaderOverhead;
//	newPacket->setByteLength(size);
//	return newPacket;
//}

void WiseBaseApplication::requestSensorReading(int index)
{
	WISE_DEBUG_3("WiseBaseApplication::requestSensorReading()\n");
	// send the request message to the Sensor Device Manager
	WiseSensorManagerMessage *reqMsg;
	reqMsg = new WiseSensorManagerMessage("application-to-sensorManager"
					      " (sample request)", 
					      SENSOR_READING_MESSAGE);
	//we need the index of the vector in the sensorTypes vector
	//to distinguish the self messages for each sensor
	reqMsg->setSensorIndex(index);
	send(reqMsg, "toSensorDeviceManager");
}

void WiseBaseApplication::toNetworkLayer(cMessage * msg)
{
	if (msg->getKind() == APPLICATION_PACKET)
		perr("toNetworkLayer() function used incorrectly to send "
		     "APPLICATION_PACKET without destination Network address");
	send(msg, "toCommunicationModule");
}

void WiseBaseApplication::toNetworkLayer(cPacket * pkt, const char *dst,
					 double delay)
{
	WiseApplicationPacket *appPkt = 
			check_and_cast <WiseApplicationPacket*>(pkt);
	appPkt->getApplicationInteractionControl().destination = string(dst);
	appPkt->getApplicationInteractionControl().source = 
						string(SELF_NETWORK_ADDRESS);
	appPkt->getApplicationInteractionControl().applicationID = 
								applicationID;
	appPkt->getApplicationInteractionControl().timestamp = simTime();
	if (delay == 0)
		send(appPkt, "toCommunicationModule");
	else
		sendDelayed(appPkt, delay, "toCommunicationModule");
}

void WiseBaseApplication::sendDirectApplicationMessage(WiseApplicationPacket *m,
						       const char *d, 
						       unsigned t)
{
	int idx = atoi(d);
	ApplicationInteractionControl_type &ctrl = 
					m->getApplicationInteractionControl();
	ctrl.source = string(SELF_NETWORK_ADDRESS);
	ctrl.destination = string(d);
	ctrl.timestamp = simTime();
	m->setType(t);
	trace() << "SEND_DIRECT_APP: to node=" << idx;
	if (idx >= 0) {
		if (idx == self) {
			// Cannot send to myself
			delete m;
			return;
		}
		send(m, "toApplicationDirect", idx);
		return;
	} 
	// To all application modules
	for (int i = 0; i < gateSize("toApplicationDirect"); i++) {
		if (i == self) // DON'T send to myself
			continue;
		send(m->dup(), "toApplicationDirect", i);
	}
	delete m;
}

void WiseBaseApplication::processAppPacket(WiseApplicationPacket *pkt)
{
	switch(pkt->getType()) {
	case WISE_APP_NORMAL:
		normalAppPacket(pkt);
		break;
	case WISE_APP_DIRECT_COMMUNICATION:
		handleDirectApplicationMessage(pkt);
		break;
	case WISE_APP_NEIGHBOUR_DISCOVER_REQUEST:
		procNeighbourDiscoverRequest(pkt);
		break;
	case WISE_APP_NEIGHBOUR_DISCOVER_REPLY:
		procNeighbourDiscoverReply(pkt);
		break;
	case WISE_APP_NEIGHBOUR_DISCOVER_TIMEOUT:
		// Should happen only on the last node
		if (!lastNode()) {
			trace() << "BAD: WISE_APP_NEIGHBOUR_DISCOVER_TIMEOUT: "
				   " should be activated only on the last node";
			return;
		}
		neighbourDiscoverCompleted();
		break;
	case WISE_APP_NEIGHBOUR_DISCOVER_DONE:
		// Should happen on all the nodes except the last
		if (lastNode()) {
			trace() << "BAD: WISE_APP_NEIGHBOUR_DISCOVER_DONE "
				   " should not be received by the last node";
			return;
		}
		notifyNeighbourDiscovered();
		break;
	default:
		trace() << "Uknown WiseApplicationPacketType = " << 
			   pkt->getType();
		break;
	}
}

void WiseBaseApplication::normalAppPacket(WiseApplicationPacket *pkt)
{
	ApplicationInteractionControl_type ctl = 
			pkt->getApplicationInteractionControl();
	if (applicationID.compare(ctl.applicationID.c_str()) != 0) 
		return;
	last_packet_latency = SIMTIME_DBL(simTime() - ctl.timestamp);
	if (latencyMax > 0 && latencyBuckets > 0) {
		double l = 1000 * last_packet_latency; // in [ms]
		collectHistogram("Application level latency [ms]", l);
	}
	fromNetworkLayer(pkt, ctl.source.c_str(), ctl.RSSI, ctl.LQI);
}
	
void WiseBaseApplication::neighbourDiscover(double startup_delay)
{
	neighbours.clear();
	if (!neighbour_discover) {
		notifyNeighbourDiscovered();
		return;
	}
	// Accounting for startup operations: DEPENDS on the protocols! 
	startup_delay += 1; 
	// Add a deterministic delay to avoid collision on 
	// neighbour discovering messages: DEPENDS on the protocols!
	// delay = node_id * base_sec ... with base_sec = 1s
	double base_sec = 1;
	startup_delay += self * base_sec; 
	if (lastNode()) {
		// Note: Only the last node has the TIMEOUT; the TIMEOUT might
		//       or might not work depending on the protocols!
		WiseApplicationPacket *tout;
		tout = new WiseApplicationPacket("Neighbour Discover TimeOut",
						 APPLICATION_PACKET);
		tout->setType(WISE_APP_NEIGHBOUR_DISCOVER_TIMEOUT);
		scheduleAt(startup_delay + (10 * base_sec), tout);
	}
	WiseApplicationPacket *m;
	m = new WiseApplicationPacket("Neighbour Discover", APPLICATION_PACKET);
	m->setType(WISE_APP_NEIGHBOUR_DISCOVER_REQUEST);
	toNetworkLayer(m, BROADCAST_NETWORK_ADDRESS, startup_delay);
	trace() << "SEND: 'Neighbour Discover Request' (broadcast) "
		   "with delay=" << startup_delay;
}

void WiseBaseApplication::procNeighbourDiscoverRequest(WiseApplicationPacket *m)
{
	ApplicationInteractionControl_type &ctrl = 
					m->getApplicationInteractionControl();
	const char *requesing_node = ctrl.source.c_str();
	trace() << "RECV: 'Neighbour Discover Request' from node=" << 
		   requesing_node;
	trace() << "      (sending reply via direct app msg)";
	m = new WiseApplicationPacket("Neighbour DiscoverED", 
				      APPLICATION_PACKET);
	sendDirectApplicationMessage(m, requesing_node, 
				     WISE_APP_NEIGHBOUR_DISCOVER_REPLY);
}

void WiseBaseApplication::procNeighbourDiscoverReply(WiseApplicationPacket *m)
{
	ApplicationInteractionControl_type &ctrl = 
					m->getApplicationInteractionControl();
	trace() << "RECV: 'Neighbour Discover Reply' from node=" << ctrl.source;
	trace() << "      (add node to neighbour list)";
	neighbours.push_back(string(ctrl.source.c_str()));
	ostringstream os;
	os << "New Neighbours Discovered!  LIST = [ ";
	vector<string>::const_iterator n;
	for (n = neighbours.begin(); n != neighbours.end(); ++n)
		os << *n << " ";
	os << "]";
	trace() << os.str();
}

void WiseBaseApplication::neighbourDiscoverCompleted()
{
	// NOTE: should be activated only on the last node!
	trace() << "Neighbours Discover COMPLETED";
	WiseApplicationPacket *m;
	m = new WiseApplicationPacket("Neighbour Discover COMPLETED", 
				      APPLICATION_PACKET);
 	// notify all the other nodes
	sendDirectApplicationMessage(m, "-1", WISE_APP_NEIGHBOUR_DISCOVER_DONE);
	// notify myself (the last node)
	notifyNeighbourDiscovered(); 		
}

void WiseBaseApplication::readParameter() 
{
	/* Get a valid references to the objects of the Resources Manager module
	 * the Mobility module and the radio module, so that we can make direct
	 * calls to their public methods.*/
	cModule *parent = getParentModule();

	if (parent->findSubmodule("ResourceManager") != -1) {
		resMgrModule = check_and_cast <ResourceManager*>
				      (parent->getSubmodule("ResourceManager"));
	} else {
		perr_app("Error in geting a valid reference to ResourceManager "
			 "for direct method calls.");
	}

	if (parent->findSubmodule("MobilityManager") != -1) {
		mobilityModule = check_and_cast <VirtualMobilityManager*>
				      (parent->getSubmodule("MobilityManager"));
	} else {
		perr_app("Error in geting a valid reference to MobilityManager "
			 "for direct method calls.");
	}
	// we make no checks here
	radioModule = check_and_cast <Radio*>
		(parent->getSubmodule("Communication")->getSubmodule("Radio"));

	self = parent->getIndex();
	if (self < 0)
		perr_app("Negative node index, very bad!");
	n_nodes = parent->getParentModule()->par("numNodes");
	cpuClockDrift = resMgrModule->getCPUClockDrift();
	setTimerDrift(cpuClockDrift);
	disabled = 1;

	applicationID = par("applicationID").stringValue();
	priority = par("priority");
	synchronizedStartup = par("synchronizedStartup");
	packetHeaderOverhead = par("packetHeaderOverhead");
	constantDataPayload = par("constantDataPayload");
	isSink = hasPar("isSink") ? par("isSink") : false;
	neighbour_discover = hasPar("neighbourDiscover") ? 
			     par("neighbourDiscover") : false;
	stringstream out;
	out << self;
	selfAddress = out.str();
}

